@@ -1,15 +1,15 @@
 /*
  * GNU Typist  - interactive typing tutor program for UNIX systems
- * 
+ *
  * Copyright (C) 1998, 1999, 2000, 2001, 2002, 2003
  *               Simon Baldwin (simonb@sco.com)
- * Copyright (C) 2003, 2004, 2008
+ * Copyright (C) 2003, 2004, 2008, 2009
  *               GNU Typist Development Team <bug-gtypist@gnu.org>
  *
- * This program is free software; you can redistribute it and/or
- * modify it under the terms of the GNU General Public License
- * as published by the Free Software Foundation, either version 3
- * of the License, or (at your option) any later version.
+ * This program is free software: you can redistribute it and/or modify
+ * it under the terms of the GNU General Public License as published by
+ * the Free Software Foundation, either version 3 of the License, or
+ * (at your option) any later version.
  *
  * This program is distributed in the hope that it will be useful,
  * but WITHOUT ANY WARRANTY; without even the implied warranty of
@@ -17,8 +17,7 @@
  * GNU General Public License for more details.
  *
  * You should have received a copy of the GNU General Public License
- * along with this program; if not, write to the Free Software
- * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
+ * along with this program.  If not, see <http://www.gnu.org/licenses/>.
  */
 
 #include "config.h"
@@ -606,6 +605,28 @@ do_instruction( FILE *script, char *line ) {
   global_prior_command = C_INSTRUCTION;
 }
 
+/*
+  Calculate whether a drill's error rate is too high, keeping in mind
+  rounding of output to a single decimal place.
+*/
+static int
+is_error_too_high( int chars_typed, int errors ) {
+  /* (double)100.0*(double)errcount / (double)total_chars ) */
+  double err_max, err;
+  char buf[BUFSIZ];
+
+  /* Calculate error rates */
+  err_max = (double)global_error_max;
+  err = (double)100.0*(double)errors / (double)chars_typed;
+
+  /* We need to use the same kind of rounding used by printf. */
+  sprintf(buf, "%.1f", err_max);
+  sscanf(buf, "%lf", &err_max);
+  sprintf(buf, "%.1f", err);
+  sscanf(buf, "%lf", &err);
+  return err > err_max;
+}
+
 /*
   execute a typing drill
 */
@@ -819,7 +840,7 @@ do_drill( FILE *script, char *line ) {
 
 	  /* check whether the error-percentage is too high (unless in d:) */
 	  if (drill_type != C_DRILL_PRACTICE_ONLY &&
-	      (float)errors/(float)chars_typed > global_error_max/100.0) 
+	      is_error_too_high(chars_typed, errors))
 	    {
 	      sprintf( message, ERROR_TOO_HIGH_MSG, global_error_max );
 	      wait_user (script, message, MODE_DRILL);
@@ -1065,7 +1086,7 @@ do_speedtest( FILE *script, char *line ) {
       
 	  /* check whether the error-percentage is too high (unless in s:) */
 	  if (drill_type != C_SPEEDTEST_PRACTICE_ONLY &&
-	      (float)errors/(float)chars_typed > global_error_max/100.0) 
+	      is_error_too_high(chars_typed, errors))
 	    {
 	      sprintf( message, ERROR_TOO_HIGH_MSG, global_error_max );
 	      wait_user (script, message, MODE_SPEEDTEST);
